/*
 * parser.c:
 *      the parser used to interprete the high
 *      level data declarations; initially just a recognizer
 *
 ******* Grammar for Declarations
 *
 *  declns = decln* tEOF
 *  decln  = tID tEQ shapes tSEMI
 *  shapes = shape (tOR shape)*
 *  shape  = tID [ tOPENBR params tCLOSEBR ]
 *  params = param (tCOMMA param)*
 *  param  = tID tID
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "bool.h"
#include "lexer.h"
#include "parser.h"

static void error( char *s )
{
	fprintf( stderr, "%s at line %d\n", s, lineno );
}

#define MUSTBE(t,mesg)	if( nexttok() != (t) ) {error(mesg); return FALSE;}

/* declns = decln* tEOF          */
BOOL parse_declns( void )
{
	while( nexttok() == tID )    /* NB: a decln starts with a tID */
	{
		ungettok();
		if( ! parse_decln() ) return FALSE;
	}
	ungettok();
	MUSTBE( tEOF, "Spurious characters found beyond EOF" );
	return TRUE;
}

/* decln  = tID tEQ shapes tSEMI */
BOOL parse_decln( void )
{
	MUSTBE( tID, "declaration expected" );
	MUSTBE( tEQ, "'=' expected" );
	if( ! parse_shapes() ) return FALSE;
	MUSTBE( tSEMI, "'or' or ';' expected" );
	return TRUE;
}

/* shapes = shape (tOR shape)* */
BOOL parse_shapes( void )
{
	do {
		if( ! parse_shape() ) return FALSE;
	} while( nexttok() == tOR );
	ungettok();
	return TRUE;
}

/* shape = tID [ tOPENBR params tCLOSEBR ] */
BOOL parse_shape( void )
{
	MUSTBE( tID, "shape name expected" );
	if( nexttok() == tOPENBR )
	{
		if( ! parse_params() ) return FALSE;
		MUSTBE( tCLOSEBR, "',' or ')' expected" );
	} else
	{
		ungettok();
	}
	return TRUE;
}

/* params = param (tCOMMA param)* */
/* param  = tID tID */
BOOL parse_params( void )
{
	do {
		MUSTBE( tID, "Field type expected" );
		MUSTBE( tID, "Field name expected" );
	} while( nexttok() == tCOMMA );
	ungettok();
	return TRUE;
}
